# Path for pylgrim result:
#   * Is an elementary simple digraph that inherits from nx.DiGraph.
#   * As it can be cyclic, it contains the information of source node.
#   * Iterator for next edge in path can be returned.
#
# Author:
#   Toon Weyens

import networkx as nx
from copy import deepcopy

class Path(nx.DiGraph):
    """Result path with source copied from graph G"""
    
    def __init__(self, G, nodes):
        super(Path, self).__init__(n_res=G.graph['n_res'])
        self.source = nodes[0]
        self.curr_node = nodes[0]
        for e_id in range(0,len(nodes)-1):
            n1, n2 = nodes[e_id:e_id+2]
            self.add_edge(n1, n2)
            for attr in G[n1][n2]:
                self[n1][n2][attr] = G[n1][n2][attr]
    
    def __str__(self):
        path_str = str(self.source)
        node = self.source
        while True:
            try:
                node = list(self.succ[node])[0]
                path_str += ' ⇨ ' + str(node)
            except IndexError:
                # last element reached
                break
            except KeyError:
                # first element had no successors
                break
        return path_str
    
    def __repr__(self):
        """identical to __str__ without the arrow."""
        path_str = str(self.source)
        node = self.source
        while True:
            try:
                node = list(self.succ[node])[0]
                path_str += ' ' + str(node)
            except IndexError:
                # last element reached
                break
            except KeyError:
                # first element had no successors
                break
        return path_str
    
    def __eq__(x,y):
        """Path considers only differences of the nodes variable."""
        return repr(x) == repr(y)
    
    def __hash__(self):
        """Path considers only differences of the nodes variable."""
        return hash(repr(self))
    
    def __iter__(self):
        # reset the counter to source when iterator is created
        self.curr_node = self.source
        return self

    def __next__(self):
        """Get next edge of path.
        Returns tuple with previous node, next node and edge attributes"""
        while True:
            prev_node = self.curr_node
            try:
                self.curr_node = list(self.succ[prev_node])[0]
                return (prev_node, self.curr_node, self[prev_node][self.curr_node])
            except IndexError:
                # last element reached
                raise StopIteration
            except KeyError:
                # first element had no successors
                raise StopIteration



# Solve Elementary Shortest Path Problem with resource constraints.
# The algorithm is based on [1].
# Features:
#   * single source / single target
#   * with resources
#   * elementary paths are sought by intelligently adding node resources to nodes that have been found to be in a NCC (negative cost cycle).
#
# Author:
#   Toon Weyens
#
# References: 
#   [1]: "Accelerated label setting algorithms for the elementary resource constrained shortest path problem" by Boland, Natashia (DOI: 10.1016/j.orl.2004.11.011)
import networkx as nx
import numpy as np
import logging
from . import tools as pt
from . import path as pth

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

# global variables
_resource_nr = 0
_resource_name = 'res_cost'

def prune_graph(G, source, target, max_res, res_name='res_cost'):
    """first step of graph {G} preprocessing
    (based on algorithm 2.1, step 0, from [1])
    Prune the graph, reducing the number of nodes and arcs, by considering least resource paths from the path {source} node to each node in the graph and from each node in the graph to the path {target} node, for each resource subject to a maximum resource in {max_res}."""
    
    global _resource_nr, _resource_name
    
    _resource_name = res_name
    
    logger.debug('Pre-process graph')
    
    # to start with, all nodes are assumed to be reachable
    n_res = G.graph['n_res']
    reachable_nodes = set(G.nodes())
    
    # iterate over all resources and delete nodes that are not reachable
    logger.debug('Delete unreachable nodes')
    for res in range(0,n_res):
        logger.debug('Treating resource {}'.format(res))
        _resource_nr = res
        
        logger.debug('Calculate feasible paths from source for resource {}'.format(res))
        lengths_source = dict(nx.single_source_dijkstra_path_length(G, source, cutoff=max_res[res], weight=_res_cost_i))
        if target not in lengths_source:
            logger.error('target not reachable for resource {}'.format(res))
            exit()
        
        logger.debug('Calculate feasible paths to target for resource {}'.format(res))
        lengths_target = dict(nx.single_source_dijkstra_path_length(G.reverse(copy=True), target, cutoff=max_res[res], weight=_res_cost_i))
        if source not in lengths_target:
            logger.error('source not reachable for resource {}'.format(res))
            exit()
        
        nodes_to_remove = set()
        for node in reachable_nodes:
            if node not in lengths_source or node not in lengths_target:
                nodes_to_remove.add(node)
            elif lengths_source[node] + lengths_target[node] > max_res[res]:
                nodes_to_remove.add(node)
        
        logger.debug('Remove {} nodes due to violation of resource'.format(len(nodes_to_remove),res))
        for node in nodes_to_remove:
            reachable_nodes.remove(node)
    
    logger.debug('{} reachable nodes:'.format(len(reachable_nodes)))
    logger.debug('      {}'.format(reachable_nodes))
    
    # set up reduced graph
    logger.debug('Set up reduced graph')
    H = nx.DiGraph(n_res=n_res)
    for node in reachable_nodes:
        for node2 in reachable_nodes:
            if G.has_edge(node, node2) and not H.has_edge(node,node2): 
                H.add_edge(node, node2)
                for attr in G[node][node2]:
                    H[node][node2][attr] = G[node][node2][attr]
    #nx.draw_circular(H,with_labels=True)
    #plt.show()
    
    # return pruned graph
    return H

def setup_least_resource_paths_ESPPRC(G, res_name='res_cost'):
    """second step of graph {G} preprocessing
    (based on algorithm 2.1, step 0, from [1])
    Solve the all-pairs shortest path problem on the graph, with lengths set to {_res_cost_i} , for each resource r."""
    
    global _resource_nr, _resource_name
    
    _resource_name = res_name
    
    # iterate over all resources and calculate least-resource paths for all pairs
    n_res = G.graph['n_res']
    logger.debug('Calculate least-resource pairs')
    res_min = list()
    for res in range(0,n_res):
        logger.debug('Treating resource {}'.format(res))
        _resource_nr = res
        res_min.append(dict(nx.all_pairs_dijkstra_path_length(G, weight=_res_cost_i)))
    
    # return preprocessed network and least-resource paths
    return res_min

def preprocess(G, source, target, max_res, res_name='res_cost'):
    """preprocess graph {G}
    (based on algorithm 2.1, step 0, from [1])"""
    
    # 1. prune graph
    H = prune_graph(G, source, target, max_res, res_name=res_name)
    
    # 2. set up least resource paths
    res_min = setup_least_resource_paths_ESPPRC(H, res_name=res_name)
    
    # return preprocessed network and least-resource paths
    return H, res_min

def GLSA(G, S, source, target, max_res, res_min, res_name='res_cost'):
    """General Label Setting Algorithm
    (based on algorithm 2.1, step 1 and 2, from [1])"""
    
    # test
    if target == source:
        logger.error('target cannot be source')
        exit()
    
    # 1. Initialization
    inf = float('inf')
    n_res = G.graph['n_res']
    # paths for each node
    paths = {source: list()}
    paths[source].append([source])
    # labels for each path (cost, resources):
    labels = {source: list()}
    labels[source].append((0,np.zeros(n_res+len(S))))
    # labels to treat: start with 0th label of path ending at source
    L = set([(source,0)])
    
    # 2. select lexicographically minimal label
    logger.debug('Loop over labels to be extended')
    while L:
        # select lexicographically minimal label
        logger.debug('Select lexicographically minimal label:')
        LML_for_prev_res= L
        for res in range(0,n_res+len(S)):
            res_LML = inf
            LML_for_this_res = []
            for label in LML_for_prev_res:
                res_loc = labels[label[0]][label[1]][1][res]
                if res_loc <= res_LML:
                    # better or equivalent label
                    res_LML = res_loc
                    LML_for_this_res.append(label)
            # set u_label to first element found
            u_label = LML_for_this_res[0]
            u = u_label[0]
            l = u_label[1]
            if len(LML_for_this_res) == 1:
                # found lexicographically minimal label
                break
            else:
                # multiple labels equivalent for this res
                LML_for_prev_res = LML_for_this_res
        L.remove(u_label)
        
        logger.debug('{}th label of node {} chosen:'.format(l,u))
        logger.debug('{} (C {} | R {})'.format(pt.print_path(paths[u][l]),labels[u][l][0],labels[u][l][1]))
        
        # extend label for each child
        for v, e in G.succ[u].items():
            logger.debug('treating edge {} -> {} (C {} | R {})'.format(u,v,e['weight'],e[res_name]))
            if len(paths.get(v,[])) > 0: logger.debug('      with current paths:')
            for n in range(0,len(paths.get(v,[]))):
                logger.debug('{} (C {} | R {})'.format(pt.print_path(paths[v][n]),labels[v][n][0],labels[v][n][1]))
            
            # error if the source is a child. The in-edges of the source need to be separated from the out-edges.
            if v == source:
                logger.critical('ERROR: source cannot be a child')
                quit()
            
            # determine whether to create a new label on the child node
            e_loc = np.zeros(n_res+len(S))
            e_loc[0:n_res] = G[u][v][res_name]
            if v in S:
                e_loc[n_res+S.index(v)] = 1
            v_label = (labels[u][l][0] +  G[u][v]['weight'], labels[u][l][1] +  e_loc)
            add_label = True
            
            # check edge resources
            for res in range(0,n_res):
                if v_label[1][res] + res_min[res][v].get(target,0.0) > max_res[res]:
                    add_label = False
                    logger.debug('at least {} more of resource {} is needed to reach target'.format(res_min[res][v].get(target,0.0),res))
                    break
            
            # check node resources
            for res in range(0,len(S)):
                if v_label[1][n_res+res] > 1:
                    add_label = False
                    logger.debug('node {} was used twice'.format(S[res]))
                    break
            
            # add
            if add_label:
                # check subset of all labels that belong to the same node for domination
                label_dominated = False
                if len(labels.get(v,[])) > 0:
                    for label in labels.get(v,[]):
                        label_dominated = _is_dominated(v_label, label)
                        if label_dominated: break
                
                if label_dominated:
                    logger.debug('but label was dominated')
                else:
                    # setup label and path
                    if labels.get(v, None) == None:
                        paths[v] = list()
                        labels[v] = list()
                    v_path = list(paths[u][l])
                    v_path.append(v)
                    
                    logger.debug('add undominated label {} (C {} | R {})'.format(pt.print_path(v_path),v_label[0],v_label[1]))
                    
                    # strong dominance: set node resource to one for nodes in S that cannot
                    # be feasibly visited with edge resources
                    for n in S:
                        for res in range(0,n_res):
                            if v_label[1][res] + res_min[res][n].get(n,0.0) > max_res[res]:
                                if v_label[1][n_res+S.index(n)] == 0:
                                    logger.debug('set strong dominance for node resource {}'.format(S.index(n)))
                                    v_label[1][n_res+S.index(n)] = 1
                    
                    # remove dominated labels
                    # Note: It is possible that two labels are identical but have different paths.
                    # The iteration therefore has to be using i_label, and not, for example, using
                    #   for label in labels.get(v,[]):
                    i_label = 0
                    n_labels = len(labels.get(v,[]))
                    while i_label < n_labels:
                        label = labels[v][i_label]
                        if _is_dominated(label, v_label):
                            logger.debug('remove dominated label {} (C {} | R {})'.format(pt.print_path(paths[v][i_label]),label[0],label[1]))
                            labels[v].pop(i_label)
                            paths[v].pop(i_label)
                            L_rename = set()
                            if (v, i_label) in L:
                                L.remove((v, i_label))
                            for i in range(i_label+1, len(labels[v])+1):
                                if (v, i) in L:
                                    L_rename.add((v,i-1))
                                    L.remove((v,i))
                            L.update(L_rename)
                            i_label -= 1
                            n_labels -= 1
                        i_label += 1
                    
                    # add label and path
                    paths[v].append(v_path)
                    labels[v].append(v_label)
                    
                    # add to list L
                    L.add((v, len(labels[v])-1))
            else:
                logger.debug('therefore do not add unfeasible label {}'.format(v_label[1]))
            
    
    # return cheapest paths with label
    logger.debug('Select cheapest path to {}'.format(target))
    
    least_cost = inf
    for p in range(0,len(labels[target])):
        if labels[target][p][0] < least_cost:
            best_path = paths[target][p]
            best_label = labels[target][p]
            least_cost = best_label[0]
    return best_path, best_label

def GSSA(G, source, target, max_res, res_min, res_name='res_cost'):
    """General State Space Augmenting Algorithm
    (based on algorithm 2.2, from [1])
    Note: The graph must have been preprocessed so that it is reduced and has the minimal resource information in {res_min}."""
    
    logger.debug('Searching for shortest path {} -> {}'.format(source, target))
    
    # initialize node resources and not done
    S = list([])
    DLA_done = False

    while not DLA_done:
        # Run dynamic labelling algorithm
        path, label = GLSA(G, S, source, target, max_res, res_min, res_name=res_name)
        logger.debug('found path {} (C {} | R {})'.format(pt.print_path(path, max_path_len_for_print=len(path)), label[0], label[1]))
        path_elems = pt.count_elems(path)
        path_max_mult = max(path_elems.values())
        if path_max_mult == 1:
            logger.debug('it is elementary')
            DLA_done = True
        else:
            logger.debug('but is it not elementary')
            node_max_mult = max(path_elems, key=path_elems.get)
            S.append(node_max_mult)
            logger.debug('Incrementing node {}, which had multiplicity {}:'.format(node_max_mult, path_max_mult))
            logger.debug('S = {}'.format(S))
        #input('PAUSED')
    
    return pth.Path(G,path), label

def _is_dominated(a, b):
    """returns whether a label {a} is dominated by another label {b}"""
    
    logger.debug('check for domination of {} by {}'.format(a,b))
    if a[0] == b[0] and all(a[1] == b[1]):
        label_dominated = False
    else:
        label_dominated = True
        if a[0] < b[0]:
            label_dominated = False
        if any(a[1] < b[1]):
            label_dominated = False
    return label_dominated

def _res_cost_i(u,v,e):
    """returns weight of certain resource with index {_resource_nr} of edge {e} from node {u} to {v}, given by variable {_resource_name}.
    It is used by all_pairs_dijkstra_path_length:
        'The weight of an edge is the value returned by the function.
        The function must accept exactly three positional arguments: the two endpoints of an edge and the dictionary of edge attributes for that edge.
        The function must return a number.'"""
    
    global _resource_nr, _resource_name
    
    return e[_resource_name][_resource_nr]



# Solve Elementary Shortest Path Problem without resource constraints.
# The algorithm is based on [1], but improved by the author.
# Features:
#   * single source / all targets
#   * no resources
#   * elementary paths are sought by requesting more and more paths for nodes which have been found to form part of a NCC (negative cost cycle), when there is no alternative.
#
# Note: The strategy in [1] is intrinsically flawed. Consider the following situation:
#   * Node 2 has a path 0->1->2.
#   * Node 3 has a path 0->3.
#   * A path from node 3 can be extended to node 2, say 0->3->2.
#   * If this path has lower cost than path 2, it replaces it.
#   * To find a path to node 3, now, by extending node 2, the only possible option is 0->3, in order not to have cycles.
#   * However, it is perfectly possible that 0->1->2->3 is better than 0->3, but 0->1->2 was thrown away.
#
# The solution to this problem is to keep more paths. The algorithm, however, then becomes very slow.
# It is therefore no advised to be used for anything but very small graphs.
#
# Author:
#   Toon Weyens
#
# References: 
#   [1]: "On the shortest path problem with negative cost cycles" by Di Puglia Pugliese, Luigi (DOI: 10.1007/s10589-015-9773-1)
from collections import deque, OrderedDict
import networkx as nx
import logging
from . import tools as pt
from . import path as pth

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

def TLAdynK(G, source, K, max_path_len = -1, retry_paths = False):
    """Truncated labelling algorithm for dynamic kSPP
    (based on algorithm 3 from [1])"""
    
    # 1. Initialization
    inf = float('inf')
    # paths for each node
    paths = {source: list()}
    paths[source].append([source])
    # costs for each path
    costs = {source: list()}
    costs[source].append(0)
    # list of nodes to treat and deque to store FIFO
    L = set([source])
    L_q = deque([source])
    # first node that has been put again in the back of the queue to retry later because it gave a NCC
    L_first_retry = None
    
    # 2. main loop for selected node
    while L_q:
        # select element FIFO
        u = L_q.popleft()
        L.remove(u)
        logger.debug('  element {}'.format(u))
        
        # Set up check for NCC:
        #   1. All requested paths K_i have been determined to node i.
        #   2. It is not possible to generate at least one elementary path.
        #   3. There is at least one NCC.
        # Test 1 is done here for all children, while the other 2 are done
        # inside the children loop.
        NCC_conds = [False]*3
        # There should not be more than K[u] costs, so == should be fine.
        NCC_conds[0] = len(costs[u]) >= K[u]
        
        # extend label for each child
        for v, e in G.succ[u].items():
            logger.debug('    treating edge {} -> {}, weight = {} with current paths:'.format(u,v,e['weight']))
            first_elem_path = 0
            for n in range(0,min(len(paths.get(v,[])),K[v])):
                logger.debug('      {} ({})'.format(pt.print_path(paths[v][n]),costs[v][n]))
            logger.debug('')
            
            # error if the source is a child. The in-edges of the source need to be separated from the out-edges.
            if v == source:
                log_str = 'ERROR: source cannot be a child'
                print(log_str)
                logger.critical(log_str)
                quit()
            
            # if first test indicates possible NCC, perform test 2 and 3
            if NCC_conds[0]:
                # Initialize test 2 to true (all paths are elementary) and then
                # check whether there is not at least one non-elementary one.
                # Also initialize test 3 to False (no NCC) and then check whether
                # there is not at least one NCC.
                NCC_conds[1] = True
                NCC_conds[2] = False
                NCCs = []
                logger.debug('      testing because test 0 was not negative')
                for ku in range(0,len(costs.get(u,[]))):
                    if not (v in paths[u][ku]):
                        logger.debug('        {} not in {}'.format(v,pt.print_path(paths[u][ku])))
                        logger.debug('          -> at least one elementary path {}'.format(ku))
                        NCC_conds[1] = False
                        first_elem_path = ku
                        break
                    else:
                        logger.debug('        {} in {}'.format(v,pt.print_path(paths[u][ku])))
                        logger.debug('        NCC criterion: {} + {} ?< {}'.format(costs[u][ku],e['weight'],costs.get(v,[inf])[0]))
                        if (costs[u][ku] + e['weight'] < costs.get(v,[inf])[0]):
                            NCC_conds[2] = True
                            for n in paths[u][ku]:
                                if not (n in NCCs):
                                    NCCs.append(n)
            
            # check for all tests
            if all(NCC_conds):
                if retry_paths and not u == L_first_retry:
                    logger.debug('      putting node {} back in the queue to retry later'.format(u))
                    
                    # put it back in queue if this is the first
                    if L_first_retry == None:
                        L_first_retry = u
                    L.add(u)
                    L_q.append(u)
                    logger.debug('  {} elements in queue'.format(len(L)))
                    
                    # skip current node
                    break
                else:
                    logger.debug('      returning {} NCC(s)'.format(len(NCCs)))
                    for i in range(0,len(NCCs)):
                        logger.debug('        {}'.format(NCCs[i]))
                    return paths, costs, NCCs
                    
            else:
                # initialize path with cost for v if not yet done
                if (costs.get(v) == None):
                    costs[v] = list()
                    paths[v] = list()
                
                # Loop over all paths of u.
                # Note that the lower boundary is possibly given by the first elementary path,
                # found in the check for NCC. The upper boundary is not taken to be equal to
                # K[u], but instead to the number of paths to that have been set up. Each of
                # these can potentially give a lower result.
                for ku in range(first_elem_path,len(paths[u])):
                    # abandon this iteration if invalid path
                    if v in paths[u][ku]:
                        continue
                    
                    logger.debug('      from node {} trying to extend path {}: {} -> {}'.format(u,ku,pt.print_path(paths[u][ku],max_path_len_for_print=3),v))
                    
                    # Loop over all paths of v.
                    # Note that even though only K[v] paths are sought ultimately, it can very well
                    # be that a number of these paths are eliminated when extending it further to other
                    # nodes, so that paths that are more costly to the current node could be needed
                    # for the extensions.
                    for kv in range(0,max(K[v],len(paths[v]))):
                        # sets up possible new path, as well as costs
                        path_v = list(paths[u][ku])
                        path_v.append(v)
                        cost_u = costs.get(u,[inf])
                        cost_v = costs.get(v,[inf])
                        cost_ku = cost_u[ku] if ku < len(cost_u) else inf
                        cost_kv = cost_v[kv] if kv < len(cost_v) else inf
                        logger.debug('        ku = {}/{}, kv = {}/{}'.format(ku,K[u]-1,kv,K[v]-1))
                        logger.debug('          current cost of path {}: {}'.format(kv,cost_kv))
                        logger.debug('          cost of potential new path = {}'.format(cost_ku+e['weight']))
                        #if (cost_ku + e['weight'] < cost_kv) or \
                            #(cost_ku + e['weight'] == cost_kv and not path_v == paths[v][kv] and not kv >= len(paths[v])):
                        # by default add path if not higher cost
                        add_new_path = cost_ku + e['weight'] <= cost_kv
                        # but check for duplicates all previous paths up to kv, as long as they have been initialized
                        for kv2 in range(0,min(kv,len(paths[v])-1)+1):
                            if path_v == paths[v][kv2]:
                                logger.debug('          path already existed in paths[{}][{}]'.format(v,kv2))
                                add_new_path = False
                                break
                        
                        # check maximum path length if positive
                        if max_path_len > 0 and len(path_v) > max_path_len:
                            logger.debug('          maximum path length {} reached'.format(max_path_len))
                            add_new_path = False
                            break
                            
                        if (add_new_path):
                            # possibly empty the variable containing first retry
                            L_first_retry = None
                            
                            # insert new path with cost
                            logger.debug('          inserting path {}({}) in path[{}] at position {}'.format(path_v,cost_ku + e['weight'],v,kv))
                            costs[v].insert(kv,cost_ku + e['weight'])
                            paths[v].insert(kv,path_v)
                            
                            # possibly add node v to L
                            if not (v in L):
                                logger.debug('          add node {} to set L'.format(v))
                                L.add(v)
                                L_q.append(v)
                            
                            # skip all following kv to next path ku
                            break
                        else:
                            logger.debug('          this path was not inserted')
            
            logger.debug('    resulting paths to {}:'.format(v))
            for n in range(0,len(paths[v])):
                logger.debug('      {}({})'.format(pt.print_path(paths[v][n]),costs[v][n]))
            logger.debug('')
        logger.debug('  {} elements in queue'.format(len(L)))
        logger.debug('')
        #print('  ------------------------------------------------------')
        #input("  Press Enter to continue...")
        #print('  ------------------------------------------------------')
        #print('')
    
    return paths, costs, []


def DLA(G, source, min_K=1, output_pos = False, max_path_len=-1):
    """Dynamic labelling algorithm
    (based on algorithm 4 from [1])"""
    
    logger.info('source: {}'.format(source))
    
    # initialize K label to minimal value and not done
    K = {}
    for n in G.nodes:
        K[n] = min_K
    DLA_done = False

    while not DLA_done:
        # Run truncated labelling algorithm for dynamic kSPP
        paths, costs, NCCs = TLAdynK(G, source, K, max_path_len)
        
        # output for tests
        logger.info('')
        logger.info('costs summary of this level:')
        costs_tot = dict()
        for n in G.nodes():
            for p in range(0,len(paths.get(n,[]))):
                costs_tot[tuple(paths[n][p])] = costs[n][p]
        
        # sort (from https://stackoverflow.com/a/15179418/3229162)
        costs_sorted = OrderedDict(sorted(costs_tot.items(), key=lambda t: t[1]))
        c_id = 0
        for c in costs_sorted:
            c_id += 1
            if costs_sorted[c] < 0 or output_pos:
                path_short = pt.print_path(c)
                logger.info('  {}: {} for {}'.format(c_id,costs_sorted[c],path_short))
        logger.info('')
        
        # check for absence negative cost cycles
        if NCCs == []:
            break
        else:
            logger.debug('updating K:')
            for n in NCCs:
                K[n] += 1
                logger.debug('  K[{}] -> {}:'.format(n,K[n]))
            logger.debug('')
            #print('========================================================')
            #input("Press Enter to continue...")
            #print('========================================================')
            #print('')
    
    # return
    rpaths = dict()
    for node in paths:
        rpaths[node] = list()
        for path in paths[node]:
            rpaths[node].append(pth.Path(G, path))
    return rpaths, costs



# Tools for pylgrim:
#   * decouple_source and undecouple_source to move all in-edges from a source node to a duplicate and vice versa.
#   * print_path to pretty print a path.
#   * count_elems to count the number of elements in a path and return a dictionary keyed with a label.
#
# Author:
#   Toon Weyens

import logging

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

def decouple_source(G, source, source_in="source_in"):
    """Decouple the source {source} of a graph {G}, by duplicating the node, called {source_in} and moving all the in-edges to it.
    Returns the number of edges displaced."""
    
    in_edges_source = tuple(G.in_edges(source))
    n_in_edges_source = len(in_edges_source)
    if n_in_edges_source > 0:
        logger.debug('displace {} in-edges ⇨ {}'.format(n_in_edges_source,source))
        
        for e in in_edges_source:
            logger.debug('  move edge {}'.format(e))
            G.add_edge(e[0],source_in)
            for attr in G[e[0]][e[1]]:
                G[e[0]][source_in][attr] = G[e[0]][e[1]][attr]
            G.remove_edge(*e)

    return n_in_edges_source


def undecouple_source(G, source, source_in="source_in"):
    """Invert the decoupling of the source {source} of a graph {G} done in decouple_source by moving all the edges to {source_in} back to {source}.
    Returns the number of edges displaced."""
    
    in_edges_source = tuple(G.in_edges(source_in))
    n_in_edges_source = len(in_edges_source)
    if n_in_edges_source > 0:
        logger.debug('place back {} in-edges {} - {}'.format(n_in_edges_source,source, in_edges_source))
        
        for e in in_edges_source:
            G.add_edge(e[0],source)
            for attr in G[e[0]][e[1]]:
                G[e[0]][source][attr] = G[e[0]][e[1]][attr]
            G.remove_edge(e[0],source_in)
        G.remove_node(source_in)

    return n_in_edges_source


def print_path(path, max_path_len_for_print = None):
    """Pretty-print a path given as an iterable of strings.
    Optionally trim if longer than max_path_len_for_print
    """
    path_short = str(path[0])
    if max_path_len_for_print is None:
        max_path_len_for_print = len(path)
    if len(path) > max_path_len_for_print + 1:
        for p in range(1,max_path_len_for_print-1):
            path_short += ' ⇨ ' + str(path[p])
        path_short += ' ⇨  …  ⇨ ' + str(path[len(path)-1])
    else:
        for p in range(1,len(path)):
            path_short += ' ⇨ ' + str(path[p])
    return path_short


def count_elems(path):
    """Count elements in a path and return a dictionary keyed with label"""
    res = dict()
    for n in path:
        res[n] = res.get(n,0)+1
    return res



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pylgrim",
    version="1.0.0",
    author="Toon Weyens",
    author_email="weyenst@gmail.com",
    description="Elementary shortest path problems, with or without resource constraints",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/ToonWeyens/pylgrim",
    packages=find_packages(),
    keywords = ['espp', 'espprc', 'shortest-path', 'graph', 'python'], # arbitrary keywords
    requires = ['networkx', 'logging', 'numpy', 'copy', 'collections', 'matplotlib', 'random', 'timeit'],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ),
)
